/*
**  CXSC is a C++ library for eXtended Scientific Computing (V 2.2)
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2006 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
**  F. Blomquist, University of Wuppertal, 25.01.2008;
*/

/* CVS $Id: lx_real.cpp,v 1.14 2006/12/06 19:14:09 cxsc Exp $ */

#include <memory.h>
#include <l_real.hpp> // Neu
#include "lx_real.hpp"
#include "lx_imath.hpp"

namespace cxsc {
	
	lx_real::lx_real(const real& n, const string &s) throw()
// Constructor:
// (10^n, string) --- > (ex,lr) vom Typ lx_real
// string s must be of the form:  
//        s = "1.234567";   or:  s = "0";
// or:    s = "1.234567e+20";
// or:    s = "1.234567e-20";
// (ex,li) is an approximation of  10^n*s;
	{
		const real c1 = 3.3219280948873623; // c1 = ln(10)/ln(2)
		const real c2 = 2.711437152598000E+015;
		const real c3 = 10000.0;
		l_real x(0);
		lx_interval u,v;
		real y,ay,r;
		int stagsave;
		bool neg;
	
		stagsave = stagprec;

		if (!Is_Integer(n))
			cxscthrow(REAL_NOT_ALLOWED(
						 "lx_real(const real&, const string& )"));
		stagprec = stagprec +1;	
		x = x + 0; // To get StagPrec(x) = actual value of stagprec;
		s >> x;    // x ~ s; x of type l_real;
		if (x==0)
		{
			lr = 0;  ex = 0;
		}
		else
		{
			neg = (n<0);
			y = n*c1;
			y = floor(_double(y));
			ay = abs(y);  r = abs(n);
			if (ay>=Max_Int_R)
				cxscthrow(REAL_NOT_ALLOWED(
							 "lx_real(const real&, const string& )"));
			u = power( lx_interval(0,l_interval(10)),r );
			v = lx_interval(ay,l_interval(1));
			if (r > c2)
			{
				v = lx_interval(ay-c3,l_interval(1));
				times2pown(u,-c3);
			}
			if (neg) u = v/u;
			else u = u/v;
			u = u*lx_interval(0,l_interval(x));
			lr = mid(li_part(u));
			r = expo(u);
			ex = add_real(y,r);
		}
		stagprec = stagsave;
		lr = adjust(lr);
	}

	std::string & operator >> (std::string &s, lx_real &a) throw()
// Writes string s to variable a of type lx_real 
// and returns an empty string s;
// Example:  s = "{-4000,2}" delivers a value a
// with:    10^(-4000)*2 ~ a;
	{
		real exr;
		int i;

		s = skipwhitespacessinglechar (s, '{'); 
		s >> exr;
		s = skipwhitespacessinglechar (s, ','); 
		i = s.find("}");
		s.erase(i+1); 
		a = lx_real(exr,s);
		s = "";

		return s;
	}

	void operator >> (const std::string &s, lx_real &a) throw()
	{
   // Writes string s to variable a of type lx_real;
		std::string r(s);
		r >> a;
	}

	void operator >> (const char *s, lx_real &a) throw()
	{
		std::string r(s);
		r >> a;
	}	
	
	bool Str_Contains(const string &s, const char &a, const char &b)
// returns 1, if s contains char a or char b, otherwise 0;
	{
		int ia,ib;
		ia = s.find(a,0);
		ib = s.find(b,0);
		return (ia>=0 || ib>=0);
	}
	
	std::istream & operator >> (std::istream &s, lx_real &a) throw()
// s must have the form: { p,r }, where r is a real value
// and the exponent p of type real must be an integer value to base 10;
// The integer value condition is tested in the constructor call:
//    a = lx_interval(exr,str);
// Blomquist, 06.101.2008; 
	{
		char c;
		real exr;
		string str,sh;
		skipeolnflag = inpdotflag = true;
		c = skipwhitespacessinglechar (s, '{');
		if (inpdotflag) s.putback(c);
		s >> SaveOpt >> exr; 
		c = skipwhitespacessinglechar (s, ',');
		if (inpdotflag) s.putback(c);
		skipeolnflag = inpdotflag = true;
		s >> str >> RestoreOpt;
		while ( !Str_Contains(str,']','}') )
		{
			c = skipwhitespaces (s);
			if (inpdotflag && c != '}') s.putback(c);
			if (c == '}') break;
			if (c == ' ') break;
			s >> sh;
			str = str + sh;
		}
		
		a = lx_real(exr,str);

		if (!waseolnflag)
		{
			skipeolnflag = false; inpdotflag = true;
			c = skipwhitespaces (s);
			if (inpdotflag && c != '}')
				s.putback(c);
		}
		return s;
	}


bool operator == (const lx_real &a, const lx_real &b) throw()
{
    const real c1 = 1e20;
    l_real ar(lr_part(a)), br(lr_part(b));
    real na(expo(a)), nb(expo(b));
    int exa(expo_gr(ar)), exb(expo_gr(br)),d;
    bool a_0(exa<-100000), b_0(exb<-100000), res;
    interval z(0);

    if (a_0 || b_0) res = (a_0 == b_0);
    else // a,b <> 0:
    {
	d = exa-exb;
	if (d>0)
	{  // br nach oben skalieren: (Overflow dabei nicht moeglich!)
	    Times2pown(br,z,d);
	    nb = nb - d;
	    if (abs(nb)>Max_Int_R)
		nb = c1;        // equality not possible
	} else // d<=0:
	{  // ar nach oben skalieren: (Overflow dabei nicht moeglich!)
	    Times2pown(ar,z,-d);
	    na = na + d;
	    if (abs(na)>Max_Int_R)
		na = c1;        // equality not possible
	}
	res = (na==nb && ar==br);
    }
    return res;
}

bool operator > (const lx_real &a, const lx_real &b) throw()
{
    l_real lra(lr_part(a)), lrb(lr_part(b));
    bool zero_a(eq_zero(a)), zero_b(eq_zero(b)), bl(false);
    real na,nb,d;
    double dbl;
    interval z(0);
    int D;

    if (zero_a) return sm_zero(b);
    if (zero_b) return gr_zero(a);
    // Now we have:  a,b != 0:
    na = expo(a);  nb = expo(b);
    d = na - nb;

    if (na==nb)
	bl = (lra >lrb);
    else 
	if (na>nb)
	{
	    if (d > 1024-expo_gr(lra))
            // Scaling not possible
		bl = (sign(lra)>0) ? true : false;
	    else
	    {
		dbl = _double(d);
		D = (int) dbl;
		Times2pown(lra,z,D);
		bl = lra > lrb;
	    }
	}
	else // nb > na
	{
	    if (d < expo_gr(lrb)-1024)
	    // Scaling not possible
		bl = (sign(lrb)<0) ? true : false;
	    else
	    {
		dbl = _double(-d);
		D = (int) dbl;
		Times2pown(lrb,z,D);
		bl = lra > lrb;
	    }
	}

    return bl;
}

void scale_up(lx_real &a) throw()
// scale_up(a) scales a.lr upwardly as far as possible. 
// Notice:    a.lr must absolutely be sorted, i.e. for
//            example: a.lr[1]=1e50, a.lr[2]=1e20, a.lr[3]=1;   
// In case of   (d>0 && a.ex-d >= -Max_Int_R)   the values 
// of 2^(a.ex)*a.lr are equal before and after the scaling.
// In case of (d<=0 || a.ex-d < -Max_Int_R) no scaling is done.
{
    interval z(0);
    int d;
    d = 1023 - expo_gr(a.lr);  // d >= 0, if a.lr <> 0;
    if (d>0 && a.ex >= -Max_Int_R+d) // If (a==0) then d<0 and nothing is done!
    {
	 	Times2pown(a.lr,z,d);    // a.lr wird mit 2^d multipliziert.
		a.ex = a.ex - d; // damit 2^(a.ex)*a.li konstant bleibt.
    }
}

void scale_down(lx_real &a) throw()
// scale_down skaliert a moeglichst weit nach unten, ohne
// dass dabei binaere Stellen verloren gehen.
// Der mathematische Wert von a bleibt dabei erhalten, d.h.
// 2^(a.ex)*a.li  hat vor und nach der Skalierung stets den
// gleichen mathematischen Wert.
{
	interval z(0);
	int d;
	d = -1021 - expo_sm(a.lr);
    // In case of a.lr=0 we get d>0, and so nothing is done! 
	if (d<0 && a.ex <= Max_Int_R+d)
	{
		Times2pown(a.lr,z,d); // a.lr = a.lr * 2^d
		a.ex = a.ex - d;
	}
}


lx_real upper_bnd(const lx_real& x) throw()
// y = upper_bnd(x) calculates an upper bound y of  x < y;
// lr = lr_part(x); (See the following code.)
// The components lr[1],lr[2], ... ,lr[StagPrec(x)] must be sorted!
{
    int stagsave = stagprec,
	p(StagPrec(x)),ex_gr;
    stagprec = p;

    lx_real res;
    l_real lr(lr_part(x));
    real ex(expo(x));

    lr = lr + 0; // sorting the lr[i]
    res = lx_real(ex,lr);
    if (p>1)
	scale_up(res);
    lr = lr_part(res);
    ex = expo(res);
    ex_gr = expo_gr(lr);
    if (ex_gr>-10000000)
    {   // x<>0:
	if (lr[1]==MaxReal)
	{   
	    times2pown(lr,-1);
	    ex = add_real(ex,1);
	}
	lr[p] = succ(lr[p]);
	lr = lr + 0; // sorting lr !
	res = lx_real(ex,lr);
    }
    else // x = 0:
    {
	lr = minreal;
	lr = adjust(lr);
	lr = lr + 0; // sorting lr !
	res = lx_real(-Max_Int_R,lr);
    }
    stagprec = stagsave;

    return res;
} // upper_bnd(...)

lx_real lower_bnd(const lx_real& x) throw()
// y = lower_bnd(x) calculates a rather great lower bound y of  x > y;
// lr = lr_part(x); (See the following code.)
// The components lr[1],lr[2], ... ,lr[StagPrec(x)] must be sorted!
{
    int stagsave = stagprec,
	p(StagPrec(x)),ex_gr;

    stagprec = p;
    lx_real res;
    l_real lr(lr_part(x));
    real ex(expo(x));

    lr = lr + 0; // sorting the lr[i]
    res = lx_real(ex,lr);
    if (p>1)
	scale_up(res);
    lr = lr_part(res);
    ex = expo(res);
    ex_gr = expo_gr(lr);
    if (ex_gr>-10000000)
    {
	if (lr[1]==-MaxReal)
	{
	    times2pown(lr,-1);
	    ex = add_real(ex,1);
	}
	lr[p] = pred(lr[p]);
	lr = lr + 0; // sorting lr !
	res = lx_real(ex,lr);
    }
    else // x = 0;
    {
	lr = -minreal;
	lr = adjust(lr);
	lr = lr + 0; // sorting lr !
	res = lx_real(-Max_Int_R,lr);
    }
    stagprec = stagsave;
    return res;
} // lower_bnd(...)

lx_real operator + (const lx_real &a, const lx_real &b) throw()
// An approximation of (a+b) is calculated. In most of the cases
// a maximum of accuracy is achieved.
{
	const real c1 = 10000.0;
	const int  c2 = 1022,
              c3 = -100000;
	int stagsave = stagprec,
	    stagmax = 39,
       exa,exb;
	real sa,sb,p,n;
	
	if (stagprec>stagmax) stagprec = stagmax;
	l_real a1,b1;
	lx_real res(0),A(a),B(b);
	
	a1 = lr_part(A);  b1 = lr_part(B);
	exa = expo_gr(a1);
	exb = expo_gr(b1);
	
	if (exa<c3) return b;  // summans a = 0;
	if (exb<c3) return a;  // summans b = 0;
	
	// From now on: a,b <> 0
	sa = add_real(exa,expo(A));
	sb = add_real(exb,expo(B));
	if (sa > sb)  // |a| >= |b|
	{
		p = c2 - exa;
		Times2pown(a1,p);
		n = sub_real(expo(A),p);
		p = sub_real(expo(B),n);
		Times2pown(b1,p);
	}
	else
	{
		p = c2 - exb;
		Times2pown(b1,p);
		n = sub_real(expo(B),p);
		p = sub_real(expo(A),n);
		Times2pown(a1,p);
	}
	a1 = a1 + b1;
	res = lx_real(n,a1);
	
	stagprec = stagsave;
	res = adjust(res);
	return res;
}

lx_real operator * (const lx_real& a, const lx_real& b) throw()
// An approximation of (a*b) is calculated. In most of the cases
// a maximum of accuracy is achieved.
{
	int stagsave = stagprec,
 stagmax = 39,
 exa,exb,d,D,k;
 real na,nb,diff;

 if (stagprec > stagmax) stagprec = stagmax;

 l_real al,bl;
 lx_real a_(a),b_(b),res(0);

 scale_down(a_);  scale_down(b_);

 al = lr_part(a_);  bl = lr_part(b_);
 exa = expo_gr(al);
 exb = expo_gr(bl);

   if (exa>-100000 && exb>-100000) // Beide Faktoren sind <> 0:
	{
		if (exa+exb <= 1022)
		{
			if (exa<0)
			{
				Times2pown(al,-exa);  
				na = add_real(expo(a_),exa);  // a = 2^na * al;
				d = 1022 - exb;
			}
			else // exa >= 0
			{
				na = expo(a_);
				d = 1022 - (exa+exb);
			}
			Times2pown(bl,d);
			nb = sub_real(expo(b_),d);  // b = 2^nb * bl;
		}
		else // exa+exb > 1022
		{
			d = exa + exb - 1022;  // d > 0;
			if (exa>exb)
			{
				D = exa - exb; // D > 0;
				if (d <= D)
				{
					Times2pown(al,-d);
					na = add_real(expo(a_),d); // a = 2^na * al;
					nb = expo(b_);
				}
				else // d > D;
				{
					k = d - D;
					if (k%2 != 0) k++;
					k = k/2;
					Times2pown(bl,-k);
					nb = add_real(expo(b_),k); // b = 2^nb * bl;
					k = k + D;
					Times2pown(al,-k);
					na = add_real(expo(a_),k); // a = 2^na * al;
				}
			}
			else // exb >= exa
			{
				D = exb - exa; // D > 0;
				if (d <= D)
				{
					Times2pown(bl,-d);
					nb = add_real(expo(b_),d); // b = 2^nb * bl;
					na = expo(a_);
				}
				else // d > D
				{
					k = d - D;
					if (k%2 != 0) k++;
					k = k/2;
					Times2pown(al,-k);
					na = add_real(expo(a_),k);  // a = 2^na * al;
					k = k + D;
					Times2pown(bl,-k);
					nb = add_real(expo(b_),k);  // b = 2^nb * bl;
				}
			}
		}
		al = al*bl;      // normal staggered multiplication

	if (na+nb<-Max_Int_R) // Underflow:
	{
		diff = (Max_Int_R + na) + nb; // diff < 0;
		Times2pown(al,diff);
		res = lx_real(-Max_Int_R,al);
	}
	else // Without underflow
	{
		na = add_real(na,nb); // Hier normale Rechnung!
		res = lx_real(na,al);
	}

	}
	stagprec = stagsave;
	res = adjust(res);
	return res;
} // operator *

lx_real operator / (const lx_real &a, const lx_real &b) throw(DIV_BY_ZERO)
{
 int stagsave = stagprec,
 stagmax = 39,
 exa,exb,d;
 real na,nb,diff;
 if (stagprec>stagmax) stagprec = stagmax;

 l_real al,bl;
 lx_real a_(a),b_(b),res;

 scale_down(a_);  
 scale_down(b_);
 al = lr_part(a_);  bl = lr_part(b_);

 exa = expo_gr(al);
 exb = expo_gr(bl);

 if (exb < -100000) 
	 cxscthrow(DIV_BY_ZERO("lx_real operator/(const lx_real &a, const lx_real &b)"));

 if (exa < -100000) return a;
   // Now it holds:  a,b <> 0:
   // First: scaling the nominator al to 1022:
 d = 1022 - exa;
 Times2pown(al,d);   // expo_gr(al) = 1022;
 na = sub_real(expo(a_),d);  // a = 2^na * al;
 if (0<=exb && exb<=511) 
	 nb = expo(b_);
 else 
	 if (exb<0)
 {  // scaling the denominator bl to 0:
	 Times2pown(bl,-exb);
	 nb = add_real(expo(b_),exb);  // b = 2^nb * bl;
 }
 else // exb > 511;
           // Now scaling the denominator bl to 511:
 {
	 d = 511 - exb;
	 Times2pown(bl,d);
	 nb = sub_real(expo(b_),d);
 }

 al = al / bl; // Normal staggered division;
    if (na-nb<-Max_Int_R) // Underflow:
	 {
		 diff = (Max_Int_R - nb) + na; // diff < 0;
		 Times2pown(al,diff);
		 res = lx_real(-Max_Int_R,al);
	 }
	 else
	 {   // Without underflow:
		 na = sub_real(na,nb); // Hier normale Rechnung!
		 res = lx_real(na,al);
	 }

	 stagprec = stagsave;
	 res = adjust(res);

	 return res;
} // operator /

l_real & l_real::operator = (const lx_real& a) throw()
{
	int exar;
	real p(expo(a)), exr;
	l_real ar(lr_part(a));
	exar = expo_gr(ar);
	if (exar<-100000) return (*this) = ar; // ar = 0;
	exr = p + exar;
	if (exr > 1024)
		cxscthrow(OVERFLOW_ERROR(
	"LX_REAL_UNREALIZABLE_AS_L_REAL: l_real & operator = (const lx_real& a)"));
	Times2pown(ar,p);
	
	return *this = ar;
}

real & real::operator = (const lx_real& a) throw()
{
	l_real lr;
	real x;
	
	lr = a;
	x = lr;
	
	return *this = x;
}

} // end: namespace cxsc
