/*
**  CXSC is a C++ library for eXtended Scientific Computing (V 2.2.3)
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2008 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* CVS $Id: xi_ari.cpp,v 1.3 2008/01/23 08:39:21 cxsc Exp $ */

//============================================================================
//
//                              Program/Module
//                                   from
//                 C++ TOOLBOX FOR VERIFIED COMPUTING I
//                         Basic Numerical Problems
//
//      Copyright (c) 1995   Rolf Hammer, Matthias Hocks, Dietmar Ratz
//
// This program/module is free software for non-commercial use. For details
// on theory, algorithms, and programs, see the book
//
//  R. Hammer, M. Hocks, U. Kulisch, D. Ratz:  C++ Toolbox for
//  Verified Computing I - Basic Numerical Problems. Springer-Verlag,
//  Heidelberg, New York, 1995.
//
// This program/module is distributed WITHOUT ANY WARRANTY. For details,
// see the "Disclaimer / Legal Matters" of the book (page iv).
//
//============================================================================
//----------------------------------------------------------------------------
// File: lxi_ari (implementation)
// Purpose: Definition of an extended interval arithmetic which allows the
//    division by an interval containing zero, based on the extended
//    staggered interval arithmetic
// Method: Overloading of operators for arithmetic and lattice operations
//    of data type 'xlx_interval'.
// Global type:
//    KindType         : component type of extended intervals
// Global function:
//    EmptyIntval()    : empty set as irregular interval
// Class xinterval:
//    operators %, -, &: operators (/,-,&) of extended interval arithmetic
//    operator =       : assignment operator
// Updates:
//    04.03.1996 : Modification of extended interval operations and kind type
//----------------------------------------------------------------------------
#include <i_util.hpp>     // Interval utility functions
#include "lxi_ari.hpp"

using namespace cxsc;
using namespace std;

//----------------------------------------------------------------------------
// An extended interval 'x', represented by the type 'xlx_interval', is 
// defined according to the following rules (a <= b):
//
// x = [a, b]              :  x.kind = Finite,     x.inf = a, x.sup = b
// x = [a, +oo]            :  x.kind = PlusInfty,  x.inf = a, x.sup undefined
// x = [-oo, a]            :  x.kind = MinusInfty, x.sup = a, x.inf undefined
// x = [-oo, a] v [b, +oo] :  x.kind = Double,     x.inf = b, x.sup = a
// x = [-oo, +oo]          :  x.kind = Double,     x.inf = a, x.sup = a
// x = [/]                 :  x.kind = Empty,      x.inf  and x.sup undefined
//
// In this definition, 'v' stands for the set union and 'oo' for infinity.
//----------------------------------------------------------------------------
interval EmptyIntval( )                      // Irregular (empty) interval
{                                             //---------------------------
  interval x;
  x = _unchecked_interval(999999999.0,-999999999.0);
  return x;
}

//----------------------------------------------------------------------------
// Constructors and assignment operator
//----------------------------------------------------------------------------
xlx_interval::xlx_interval ( )
{
  kind = Finite;
  inf  = 0.0;
  sup  = 0.0;
}

xlx_interval::xlx_interval ( const KindType& k, 
									  const lx_real& i, const lx_real& s )
{
  kind = k;
  inf  = i;
  sup  = s;
}

xlx_interval::xlx_interval ( const xlx_interval& a )
{
  kind = a.kind;
  inf  = a.inf;
  sup  = a.sup;
}

xlx_interval& xlx_interval::operator= ( const xlx_interval& a )
{
  kind = a.kind;
  inf  = a.inf;
  sup  = a.sup;
  return *this;
}

//----------------------------------------------------------------------------
// Extended interval division 'A / B' where 0 in 'B' is allowed.
//----------------------------------------------------------------------------
xlx_interval operator% ( const lx_interval& A, const lx_interval& B )
{
  lx_interval  c;
  xlx_interval Q;

  if ( in(0.0, B) ) {
    if ( in(0.0, A) ) {
      Q.kind = Double;                    // Q = [-oo,+oo] = [-oo,0] v [0,+oo]
      Q.sup  = 0.0;                       //----------------------------------
      Q.inf  = 0.0;
    }
    else if ( B == 0.0 ) {                                          // Q = [/]
      Q.kind = PlusInfty;                                           //--------
      // Q.inf  = divd(Sup(A),Inf(B));
		c = lx_interval(Sup(A))/lx_interval(Inf(B)); // blom
		Q.inf = Inf(c);                              // blom
    }
    else if ( (Sup(A) < 0.0) && (Sup(B) == 0.0) ) {         // Q = [Q.inf,+oo]
      Q.kind = PlusInfty;                                   //----------------
      // Q.inf  = divd(Sup(A),Inf(B));
		c = lx_interval(Sup(A))/lx_interval(Inf(B)); // blom
		Q.inf = Inf(c);                              // blom
    }
    else if ( (Sup(A) < 0.0) && (Inf(B) < 0.0) && (Sup(B) > 0.0) ) {
      Q.kind = Double;                        // Q = [-oo,Q.sup] v [Q.inf,+oo]
      // Q.sup  = divu(Sup(A),Sup(B));        //------------------------------
		c = lx_interval(Sup(A))/lx_interval(Sup(B)); // blom
		Q.sup = Sup(c);                              // blom
      // Q.inf  = divd(Sup(A),Inf(B));
		c = lx_interval(Sup(A))/lx_interval(Inf(B)); // blom
		Q.inf = Inf(c);                              // blom
    }
    else if ( (Sup(A) < 0.0) && (Inf(B) == 0.0) ) {         // Q = [-oo,Q.sup]
      Q.kind = MinusInfty;                                  //----------------
      // Q.sup  = divu(Sup(A),Sup(B));
		c = lx_interval(Sup(A))/lx_interval(Sup(B)); // blom
		Q.sup = Sup(c);                              // blom
    }
    else if ( (Inf(A) > 0.0) && (Sup(B) == 0.0) ) {         // Q = [-oo,Q.sup]
      Q.kind = MinusInfty;                                  //----------------
      // Q.sup  = divu(Inf(A),Inf(B));
		c = lx_interval(Inf(A))/lx_interval(Inf(B)); // blom
		Q.sup = Sup(c);                              // blom
    }
    else if ( (Inf(A) > 0.0) && (Inf(B) < 0.0) && (Sup(B) > 0.0) ) {
      Q.kind = Double;                        // Q = [-oo,Q.sup] v [Q.inf,+oo]
      // Q.sup  = divu(Inf(A),Inf(B));        //------------------------------
		c = lx_interval(Inf(A))/lx_interval(Inf(B)); // blom
		Q.sup = Sup(c);                              // blom
      // Q.inf  = divd(Inf(A),Sup(B));
		c = lx_interval(Inf(A))/lx_interval(Sup(B)); // blom
		Q.inf = Inf(c);                              // blom
    }
    else { // if ( (Inf(A) > 0.0) && (Inf(B) == 0.0) )
      Q.kind = PlusInfty;                                   // Q = [Q.inf,+oo]
      // Q.inf  = divd(Inf(A),Sup(B));                      //----------------
		c = lx_interval(Inf(A))/lx_interval(Sup(B)); // blom
		Q.inf = Inf(c);                              // blom
    }
  } // in(0.0,B)
  else {  // !in(0.0,B)
    c = A / B;                                            // Q = [C.inf,C.sup]
    Q.kind = Finite;                                      //------------------
    Q.inf  = Inf(c);
    Q.sup  = Sup(c);
  }

  return Q;
} // operator%

//----------------------------------------------------------------------------
// Subtraction of an extended interval 'B' from a real value 'a'.
//----------------------------------------------------------------------------
xlx_interval operator- ( const lx_real& a, const xlx_interval& B )
{
  xlx_interval D;
  lx_interval c;  // blom

  switch (B.kind) {
    case Finite     : D.kind = Finite;                    // D = [D.inf,D.sup]
                      // D.inf  = subd(a,B.sup);          //------------------
							 c = lx_interval(a) - lx_interval(B.sup); // blom
							 D.inf = Inf(c);                          // blom
                      // D.sup  = subu(a,B.inf);
							 c = lx_interval(a) - lx_interval(B.inf); // blom
							 D.sup = Sup(c);                          // blom
                      break;
    case PlusInfty  : D.kind = MinusInfty;                    // D = [inf,+oo]
                      // D.sup  = subu(a,B.inf);              //--------------
	                   c = lx_interval(a) - lx_interval(B.inf); // blom
	                   D.sup = Sup(c);                          // blom
                      break;
    case MinusInfty : D.kind = PlusInfty;                     // D = [-oo,sup]
                      // D.inf  = subd(a,B.sup);              //--------------
	                   c = lx_interval(a) - lx_interval(B.sup); // blom
	                   D.inf = Inf(c);
                      break;
    case Double     : D.kind = Double;        // D = [-oo,D.sup] v [D.inf,+oo]
                      // D.inf  = subd(a,B.sup); //---------------------------
	                   c = lx_interval(a) - lx_interval(B.sup); // blom
	                   D.inf = Inf(c);                          // blom
                      // D.sup  = subu(a,B.inf);
							 c = lx_interval(a) - lx_interval(B.inf); // blom
							 D.sup = Sup(c);
                      if (D.inf < D.sup) D.inf = D.sup;
                      break;
    case Empty      : D.kind = Empty;                               // D = [/]
                      // D.inf  = subd(a,B.sup);                    //--------
	                   c = lx_interval(a) - lx_interval(B.sup); // blom
	                   D.inf = Inf(c);
                      break;
  } // switch
  return D;
}

//----------------------------------------------------------------------------
// Intersection of an interval 'X' and an extended interval 'Y'. The result
// is given as a pair (vector) of intervals, where one or both of them can
// be empty intervals.
//----------------------------------------------------------------------------
lx_ivector operator& ( const lx_interval& X, const xlx_interval& Y )
{
  lx_interval H;
  lx_ivector  IS(2);

  IS[1] = EmptyIntval();
  IS[2] = EmptyIntval();

  switch (Y.kind) {
    case Finite     : // [X.inf,X.sup] & [Y.inf,Y.sup]
                      //------------------------------
                      H = lx_interval(Y.inf,Y.sup);
                      if ( !Disjoint(X,H) ) IS[1] = X & H;
                      break;
    case PlusInfty  : // [X.inf,X.sup] & [Y.inf,+oo]
                      //----------------------------
                      if (Sup(X) >= Y.inf)
                        if (Inf(X) > Y.inf)
                          IS[1] = X;
                        else
                          IS[1] = lx_interval(Y.inf,Sup(X));
                      break;
    case MinusInfty : // [X.inf,X.sup] & [-oo,Y.sup]
                      //----------------------------
                      if (Y.sup >= Inf(X))
                        if (Sup(X)<Y.sup)
                          IS[1] = X;
                        else
                          IS[1] = lx_interval(Inf(X),Y.sup);
                      break;
    case Double     : if ( (Inf(X) <= Y.sup) && (Y.inf <= Sup(X)) ) {
                        IS[1] = lx_interval(Inf(X),Y.sup);    // X & [-oo,Y.sup]
                        IS[2] = lx_interval(Y.inf,Sup(X));    // X & [Y.inf,+oo]
                      }
                      else if (Y.inf <= Sup(X)) // [X.inf,X.sup] & [Y.inf,+oo]
                        if (Inf(X) >= Y.inf)    //----------------------------
                          IS[1] = X;
                        else
                          IS[1] = lx_interval(Y.inf,Sup(X));
                      else if (Inf(X) <= Y.sup) // [X.inf,X.sup] & [-oo,Y.sup]
                        if (Sup(X) <= Y.sup)    //----------------------------
                          IS[1] = X;
                        else
                          IS[1] = lx_interval(Inf(X),Y.sup);
                      break;
    case Empty      : break;                           // [X.inf,X.sup] ** [/]
  } // switch                                          //---------------------

  return IS;
} // operator&

