/*
**  Copyright (C) 1999-2006 F. Blomquist, M. Braeuer,
**                          W. Hofschuster, W. Kraemer
**                          Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

////////////////////////////////////////////////////////////////////////////
//
//  Headerfile lxitaylor.hpp for onedimensional extended staggered 
//  Taylor-arithmetic; Definition of class lx_itaylor;
//
////////////////////////////////////////////////////////////////////////////

/*--------------------------------------------------------------------------

Compiler: g++-4.2

Definition of the class l_itaylor:

class lx_itaylor for calculating all Taylor coefficients up to the 
maximal order p.

  Elements:
            int p .................... maximal order of the Taylor coefficients
            lx_ivector tayl ............. lx_interval vector; Lb=0, Ub=p;

	    static l_ivector faks ........ storing n! with n<=170
            static int initialized ....... switcher for initialization of faks.
            static void initialize()...... performs initialization of faks.
	    explicit l_itaylor(int order). no automatic type matching

  Element functions, methodes:

            look the implementation

Implementation file: lxitaylor.cpp

-----------------------------------------------------------------------*/

#ifndef LXITAYLOR_H
#define LXITAYLOR_H

// cxsc headers
#include "lx_imath.hpp"
#include "lx_interval.hpp"
#include "lx_ivector.hpp"
#include <idot.hpp>

//C++ standard headers
#include <iostream>

using namespace cxsc;

namespace taylor{
enum{
    _i_ln,

    _i_tan,
    _i_cot,

    _i_asin,
    _i_acos,
    _i_atan,
    _i_acot,

    _i_tanh,
    _i_coth,

    _i_asinh,
    _i_acosh,
    _i_atanh,
    _i_acoth,
};

///////////////////////////////////////////////////////////////
//
//                   Class lx_itaylor
//
///////////////////////////////////////////////////////////////

// Class lx_itaylor for calculating all Taylor-coefficients up to order p
// of functions with one independent variable.

class lx_itaylor;

typedef lx_itaylor (*df_FctPtr)(lx_itaylor);

class lx_itaylor{

 private:
  int p;         // max. Taylor-order of the object;
  lx_ivector tayl; // Lx_interval vector with Taylor-coefficients

  static lx_ivector faks;
  static int initialized;
  static void initialize();
  explicit lx_itaylor(int order);

 public:
  // Constructors and Destructors:
  lx_itaylor();
  lx_itaylor(const lx_itaylor& );
  lx_itaylor(int order, const real&   value);      // (x,1,0,...,0)
  lx_itaylor(int order, const l_real& value);      // (x,1,0,...,0)
  lx_itaylor(int order, const lx_real& value);     // (x,1,0,...,0)
  lx_itaylor(int order, const interval& value);    // (x,1,0,...,0)
  lx_itaylor(int order, const l_interval& value);  // (x,1,0,...,0)
  lx_itaylor(int order, const lx_interval& value); // (x,1,0,...,0)

  ~lx_itaylor(){;};

  // Initialization functions for independent variables (x,1,0,...,0):
  // Caution: (x,1,0,...,0)  for real x conversion errors are possible!
  lx_itaylor var_lx_itaylor(int ord, const real& c);
  lx_itaylor var_lx_itaylor(int ord, const l_real& c);
  lx_itaylor var_lx_itaylor(int ord, const lx_real& c);
  lx_itaylor var_lx_itaylor(int ord, const interval& c);
  lx_itaylor var_lx_itaylor(int ord, const l_interval& c);
  lx_itaylor var_lx_itaylor(int ord, const lx_interval& c);

  // Initialization functions for constants (c,0,0,...,0):
  // Caution: (c,0,0,...,0)  for real c conversion errors are possible!
  friend lx_itaylor const_lx_itaylor(int ord, const real& c);
  friend lx_itaylor const_lx_itaylor(int ord, const l_real& c);
  friend lx_itaylor const_lx_itaylor(int ord, const lx_real& c);
  friend lx_itaylor const_lx_itaylor(int ord, const interval& c);
  friend lx_itaylor const_lx_itaylor(int ord, const l_interval& c);
  friend lx_itaylor const_lx_itaylor(int ord, const lx_interval& c);
  
  // assignment operators:
  lx_itaylor operator=(const lx_itaylor& );
  lx_itaylor operator=(int);
  lx_itaylor operator=(const real& );
  lx_itaylor operator=(const l_real& );
  lx_itaylor operator=(const lx_real& );
  lx_itaylor operator=(const interval& );
  lx_itaylor operator=(const l_interval& ); 
  lx_itaylor operator=(const lx_interval& ); 

  // access to the components:
  friend int get_order(const lx_itaylor& x);
  friend lx_ivector get_all_coef(const lx_itaylor& x);
  friend lx_interval get_j_coef(const lx_itaylor& x, int j);

  // access to the derivative of order j:
  friend lx_interval get_j_derivative(const lx_itaylor& x, int j);

  // Output:
  friend void print_base10_lx_itaylor(const lx_itaylor& x);
  friend void print_base2_lx_itaylor(const lx_itaylor& x);

  // Overloading the operators for elements of the class l_itaylor:

  // operator - :
  friend lx_itaylor operator-(const lx_itaylor& x);

  // operators +,-,*,/  for (lx_itaylor, lx_itaylor):
  friend lx_itaylor operator-(const lx_itaylor& x, const lx_itaylor& y);
  friend lx_itaylor operator+(const lx_itaylor& x, const lx_itaylor& y);
  friend lx_itaylor operator*(const lx_itaylor& x, const lx_itaylor& y);
  friend lx_itaylor operator/(const lx_itaylor& x, const lx_itaylor& y);

  // operators +,-,*,/ for (interval, lx_itaylor):
  friend lx_itaylor operator-(const interval& x, const lx_itaylor& y);
  friend lx_itaylor operator+(const interval& x, const lx_itaylor& y);
  friend lx_itaylor operator*(const interval& x, const lx_itaylor& y);
  friend lx_itaylor operator/(const interval& x, const lx_itaylor& y);

  // operators +,-,*,/ for (l_interval, lx_itaylor):
  friend lx_itaylor operator-(const l_interval& x, const lx_itaylor& y);
  friend lx_itaylor operator+(const l_interval& x, const lx_itaylor& y);
  friend lx_itaylor operator*(const l_interval& x, const lx_itaylor& y);
  friend lx_itaylor operator/(const l_interval& x, const lx_itaylor& y);
  
    // operators +,-,*,/ for (lx_interval, lx_itaylor):
  friend lx_itaylor operator-(const lx_interval& x, const lx_itaylor& y);
  friend lx_itaylor operator+(const lx_interval& x, const lx_itaylor& y);
  friend lx_itaylor operator*(const lx_interval& x, const lx_itaylor& y);
  friend lx_itaylor operator/(const lx_interval& x, const lx_itaylor& y);

  // operators +,-,*,/ for (lx_itaylor, interval):
  friend lx_itaylor operator-(const lx_itaylor& x, const interval& y);
  friend lx_itaylor operator+(const lx_itaylor& x, const interval& y);
  friend lx_itaylor operator*(const lx_itaylor& x, const interval& y);
  friend lx_itaylor operator/(const lx_itaylor& x, const interval& y);

  // operators +,-,*,/ for (lx_itaylor, l_interval):
  friend lx_itaylor operator-(const lx_itaylor& x, const l_interval& y);
  friend lx_itaylor operator+(const lx_itaylor& x, const l_interval& y);
  friend lx_itaylor operator*(const lx_itaylor& x, const l_interval& y);
  friend lx_itaylor operator/(const lx_itaylor& x, const l_interval& y);
  
    // operators +,-,*,/ for (lx_itaylor, lx_interval):
  friend lx_itaylor operator-(const lx_itaylor& x, const lx_interval& y);
  friend lx_itaylor operator+(const lx_itaylor& x, const lx_interval& y);
  friend lx_itaylor operator*(const lx_itaylor& x, const lx_interval& y);
  friend lx_itaylor operator/(const lx_itaylor& x, const lx_interval& y);

  // operators +,-,*,/ for (real, lx_itaylor):
  friend lx_itaylor operator-(const real& x, const lx_itaylor& y);
  friend lx_itaylor operator+(const real& x, const lx_itaylor& y);
  friend lx_itaylor operator*(const real& x, const lx_itaylor& y);
  friend lx_itaylor operator/(const real& x, const lx_itaylor& y);

  // operators +,-,*,/ for (int, lx_itaylor):
  friend lx_itaylor operator-(int x, const lx_itaylor& y);
  friend lx_itaylor operator+(int x, const lx_itaylor& y);
  friend lx_itaylor operator*(int x, const lx_itaylor& y);
  friend lx_itaylor operator/(int x, const lx_itaylor& y);

  // operators +,-,*,/ for (l_real, lx_itaylor):
  friend lx_itaylor operator-(const l_real& x, const lx_itaylor& y);
  friend lx_itaylor operator+(const l_real& x, const lx_itaylor& y);
  friend lx_itaylor operator*(const l_real& x, const lx_itaylor& y);
  friend lx_itaylor operator/(const l_real& x, const lx_itaylor& y);

  // operators +,-,*,/ for (lx_real, lx_itaylor):
  friend lx_itaylor operator-(const lx_real& x, const lx_itaylor& y);
  friend lx_itaylor operator+(const lx_real& x, const lx_itaylor& y);
  friend lx_itaylor operator*(const lx_real& x, const lx_itaylor& y);
  friend lx_itaylor operator/(const lx_real& x, const lx_itaylor& y);

  // operators +,-,*,/ for (lx_itaylor, real):
  friend lx_itaylor operator-(const lx_itaylor& x, const real& y);
  friend lx_itaylor operator+(const lx_itaylor& x, const real& y);
  friend lx_itaylor operator*(const lx_itaylor& x, const real& y);
  friend lx_itaylor operator/(const lx_itaylor& x, const real& y);

  // operators +,-,*,/ for (lx_itaylor, int):
  friend lx_itaylor operator-(const lx_itaylor& x, int y);
  friend lx_itaylor operator+(const lx_itaylor& x, int y);
  friend lx_itaylor operator*(const lx_itaylor& x, int y);
  friend lx_itaylor operator/(const lx_itaylor& x, int y);

  // operators +,-,*,/ for (lx_itaylor, l_real):
  friend lx_itaylor operator-(const lx_itaylor& x, const l_real& y);
  friend lx_itaylor operator+(const lx_itaylor& x, const l_real& y);
  friend lx_itaylor operator*(const lx_itaylor& x, const l_real& y);
  friend lx_itaylor operator/(const lx_itaylor& x, const l_real& y);
  
  // operators +,-,*,/ for (lx_itaylor, lx_real):
  friend lx_itaylor operator-(const lx_itaylor& x, const lx_real& y);
  friend lx_itaylor operator+(const lx_itaylor& x, const lx_real& y);
  friend lx_itaylor operator*(const lx_itaylor& x, const lx_real& y);
  friend lx_itaylor operator/(const lx_itaylor& x, const lx_real& y);


  // Overloading the standard functions:
  friend lx_itaylor sqr(const lx_itaylor& x);
  friend lx_itaylor sqrt(const lx_itaylor& x);
  friend lx_itaylor sqrt(const lx_itaylor& x, int n);
  friend lx_itaylor sqrt1px2(const lx_itaylor& x);
  friend lx_itaylor sqrtp1m1(const lx_itaylor& x);
  friend lx_itaylor sqrt1mx2(const lx_itaylor& x);
  friend lx_itaylor sqrtx2m1(const lx_itaylor& x);

  friend lx_itaylor pow(const lx_itaylor& x, const lx_interval& alpha);
  friend lx_itaylor power(const lx_itaylor& x, int n);

  friend lx_itaylor exp(const lx_itaylor& x);
  friend lx_itaylor exp2(const lx_itaylor& x);
  friend lx_itaylor exp10(const lx_itaylor& x);
  friend lx_itaylor expm1(const lx_itaylor& x);

  // Help function
  friend void f_g_u(lx_itaylor& f, const lx_itaylor& g, 
                    const lx_itaylor& u, int nb_function);

  friend lx_itaylor ln(const lx_itaylor& x);
  friend lx_itaylor log2(const lx_itaylor& x);
  friend lx_itaylor log10(const lx_itaylor& x);
  friend lx_itaylor lnp1(const lx_itaylor& x);

  friend lx_itaylor sin(const lx_itaylor& x);
  friend lx_itaylor cos(const lx_itaylor& x);
  friend lx_itaylor tan(const lx_itaylor& x);
  friend lx_itaylor cot(const lx_itaylor& x);

  friend lx_itaylor sinh(const lx_itaylor& x);
  friend lx_itaylor cosh(const lx_itaylor& x);
  friend lx_itaylor tanh(const lx_itaylor& x);
  friend lx_itaylor coth(const lx_itaylor& x);

  friend lx_itaylor asin(const lx_itaylor& x);
  friend lx_itaylor acos(const lx_itaylor& x);
  friend lx_itaylor atan(const lx_itaylor& x);
  friend lx_itaylor acot(const lx_itaylor& x);

  friend lx_itaylor asinh(const lx_itaylor& x);
  friend lx_itaylor acosh(const lx_itaylor& x);
  friend lx_itaylor atanh(const lx_itaylor& x);
  friend lx_itaylor acoth(const lx_itaylor& x);
  
  friend void fEval  ( df_FctPtr, lx_interval, lx_interval& );
  friend void dfEval ( df_FctPtr, lx_interval, lx_interval&, 
							                          lx_interval& );
}; // End of class lx_itaylor

 lx_itaylor const_lx_itaylor(int ord, const real& c);
 lx_itaylor const_lx_itaylor(int ord, const l_real& c);
 lx_itaylor const_lx_itaylor(int ord, const lx_real& c);
 lx_itaylor const_lx_itaylor(int ord, const interval& c);
 lx_itaylor const_lx_itaylor(int ord, const l_interval& c);
 lx_itaylor const_lx_itaylor(int ord, const lx_interval& c);
  
 int get_order(const lx_itaylor& x);
 lx_ivector get_all_coef(const lx_itaylor& x);
 lx_interval get_j_coef(const lx_itaylor& x, int j);

  // access to the derivative of order j:
 lx_interval get_j_derivative(const lx_itaylor& x, int j);

  // Output:
 void print_base10_lx_itaylor(const lx_itaylor& x);
 void print_base2_lx_itaylor(const lx_itaylor& x);

  // Overloading the operators for elements of the class l_itaylor:

  // operator - :
 lx_itaylor operator-(const lx_itaylor& x);

  // operators +,-,*,/  for (lx_itaylor, lx_itaylor):
 lx_itaylor operator-(const lx_itaylor& x, const lx_itaylor& y);
 lx_itaylor operator+(const lx_itaylor& x, const lx_itaylor& y);
 lx_itaylor operator*(const lx_itaylor& x, const lx_itaylor& y);
 lx_itaylor operator/(const lx_itaylor& x, const lx_itaylor& y);

  // operators +,-,*,/ for (interval, lx_itaylor):
 lx_itaylor operator-(const interval& x, const lx_itaylor& y);
 lx_itaylor operator+(const interval& x, const lx_itaylor& y);
 lx_itaylor operator*(const interval& x, const lx_itaylor& y);
 lx_itaylor operator/(const interval& x, const lx_itaylor& y);

  // operators +,-,*,/ for (l_interval, lx_itaylor):
 lx_itaylor operator-(const l_interval& x, const lx_itaylor& y);
 lx_itaylor operator+(const l_interval& x, const lx_itaylor& y);
 lx_itaylor operator*(const l_interval& x, const lx_itaylor& y);
 lx_itaylor operator/(const l_interval& x, const lx_itaylor& y);
  
    // operators +,-,*,/ for (lx_interval, lx_itaylor):
 lx_itaylor operator-(const lx_interval& x, const lx_itaylor& y);
 lx_itaylor operator+(const lx_interval& x, const lx_itaylor& y);
 lx_itaylor operator*(const lx_interval& x, const lx_itaylor& y);
 lx_itaylor operator/(const lx_interval& x, const lx_itaylor& y);

  // operators +,-,*,/ for (lx_itaylor, interval):
 lx_itaylor operator-(const lx_itaylor& x, const interval& y);
 lx_itaylor operator+(const lx_itaylor& x, const interval& y);
 lx_itaylor operator*(const lx_itaylor& x, const interval& y);
 lx_itaylor operator/(const lx_itaylor& x, const interval& y);

  // operators +,-,*,/ for (lx_itaylor, l_interval):
 lx_itaylor operator-(const lx_itaylor& x, const l_interval& y);
 lx_itaylor operator+(const lx_itaylor& x, const l_interval& y);
 lx_itaylor operator*(const lx_itaylor& x, const l_interval& y);
 lx_itaylor operator/(const lx_itaylor& x, const l_interval& y);
  
    // operators +,-,*,/ for (lx_itaylor, lx_interval):
 lx_itaylor operator-(const lx_itaylor& x, const lx_interval& y);
 lx_itaylor operator+(const lx_itaylor& x, const lx_interval& y);
 lx_itaylor operator*(const lx_itaylor& x, const lx_interval& y);
 lx_itaylor operator/(const lx_itaylor& x, const lx_interval& y);

  // operators +,-,*,/ for (real, lx_itaylor):
 lx_itaylor operator-(const real& x, const lx_itaylor& y);
 lx_itaylor operator+(const real& x, const lx_itaylor& y);
 lx_itaylor operator*(const real& x, const lx_itaylor& y);
 lx_itaylor operator/(const real& x, const lx_itaylor& y);

  // operators +,-,*,/ for (int, lx_itaylor):
 lx_itaylor operator-(int x, const lx_itaylor& y);
 lx_itaylor operator+(int x, const lx_itaylor& y);
 lx_itaylor operator*(int x, const lx_itaylor& y);
 lx_itaylor operator/(int x, const lx_itaylor& y);

  // operators +,-,*,/ for (l_real, lx_itaylor):
 lx_itaylor operator-(const l_real& x, const lx_itaylor& y);
 lx_itaylor operator+(const l_real& x, const lx_itaylor& y);
 lx_itaylor operator*(const l_real& x, const lx_itaylor& y);
 lx_itaylor operator/(const l_real& x, const lx_itaylor& y);

  // operators +,-,*,/ for (lx_real, lx_itaylor):
 lx_itaylor operator-(const lx_real& x, const lx_itaylor& y);
 lx_itaylor operator+(const lx_real& x, const lx_itaylor& y);
 lx_itaylor operator*(const lx_real& x, const lx_itaylor& y);
 lx_itaylor operator/(const lx_real& x, const lx_itaylor& y);

  // operators +,-,*,/ for (lx_itaylor, real):
 lx_itaylor operator-(const lx_itaylor& x, const real& y);
 lx_itaylor operator+(const lx_itaylor& x, const real& y);
 lx_itaylor operator*(const lx_itaylor& x, const real& y);
 lx_itaylor operator/(const lx_itaylor& x, const real& y);

  // operators +,-,*,/ for (lx_itaylor, int):
 lx_itaylor operator-(const lx_itaylor& x, int y);
 lx_itaylor operator+(const lx_itaylor& x, int y);
 lx_itaylor operator*(const lx_itaylor& x, int y);
 lx_itaylor operator/(const lx_itaylor& x, int y);

  // operators +,-,*,/ for (lx_itaylor, l_real):
 lx_itaylor operator-(const lx_itaylor& x, const l_real& y);
 lx_itaylor operator+(const lx_itaylor& x, const l_real& y);
 lx_itaylor operator*(const lx_itaylor& x, const l_real& y);
 lx_itaylor operator/(const lx_itaylor& x, const l_real& y);
  
  // operators +,-,*,/ for (lx_itaylor, lx_real):
 lx_itaylor operator-(const lx_itaylor& x, const lx_real& y);
 lx_itaylor operator+(const lx_itaylor& x, const lx_real& y);
 lx_itaylor operator*(const lx_itaylor& x, const lx_real& y);
 lx_itaylor operator/(const lx_itaylor& x, const lx_real& y);


  // Overloading the standard functions:
 lx_itaylor sqr(const lx_itaylor& x);
 lx_itaylor sqrt(const lx_itaylor& x);
 lx_itaylor sqrt(const lx_itaylor& x, int n);
 lx_itaylor sqrt1px2(const lx_itaylor& x);
 lx_itaylor sqrtp1m1(const lx_itaylor& x);
 lx_itaylor sqrt1mx2(const lx_itaylor& x);
 lx_itaylor sqrtx2m1(const lx_itaylor& x);

 lx_itaylor pow(const lx_itaylor& x, const lx_interval& alpha);
 lx_itaylor power(const lx_itaylor& x, int n);

 lx_itaylor exp(const lx_itaylor& x);
 lx_itaylor exp2(const lx_itaylor& x);
 lx_itaylor exp10(const lx_itaylor& x);
 lx_itaylor expm1(const lx_itaylor& x);

  // Help function
 void f_g_u(lx_itaylor& f, const lx_itaylor& g, 
						  const lx_itaylor& u, int nb_function);

 lx_itaylor ln(const lx_itaylor& x);
 lx_itaylor log2(const lx_itaylor& x);
 lx_itaylor log10(const lx_itaylor& x);
 lx_itaylor lnp1(const lx_itaylor& x);

 lx_itaylor sin(const lx_itaylor& x);
 lx_itaylor cos(const lx_itaylor& x);
 lx_itaylor tan(const lx_itaylor& x);
 lx_itaylor cot(const lx_itaylor& x);

 lx_itaylor sinh(const lx_itaylor& x);
 lx_itaylor cosh(const lx_itaylor& x);
 lx_itaylor tanh(const lx_itaylor& x);
 lx_itaylor coth(const lx_itaylor& x);

 lx_itaylor asin(const lx_itaylor& x);
 lx_itaylor acos(const lx_itaylor& x);
 lx_itaylor atan(const lx_itaylor& x);
 lx_itaylor acot(const lx_itaylor& x);

 lx_itaylor asinh(const lx_itaylor& x);
 lx_itaylor acosh(const lx_itaylor& x);
 lx_itaylor atanh(const lx_itaylor& x);
 lx_itaylor acoth(const lx_itaylor& x);

 void fEval  ( df_FctPtr, lx_interval, lx_interval& );
 void dfEval ( df_FctPtr, lx_interval, lx_interval&, lx_interval& );
} // End of namespace taylor

#endif
